@@ -14,73 +14,12 @@ import serial, Queue, select
 import select
 
 # allow running without installing
-sys.path.insert(0, os.path.join(os.path.dirname(os.path.realpath(__file__)), '..'))
+#sys.path.append(os.path.join(os.path.dirname(os.path.realpath(__file__)), '..'))
 
 
 from MAVProxy.modules.lib import textconsole
 from MAVProxy.modules.lib import mp_settings
 
-class MPSettings(object):
-    def __init__(self):
-        self.vars = [ ('link', int),
-                      ('altreadout', int),
-                      ('distreadout', int),
-                      ('battreadout', int),
-                      ('heartbeat', int),
-                      ('numcells', int),
-                      ('speech', int),
-                      ('mavfwd', int),
-                      ('mavfwd_rate', int),
-                      ('streamrate', int),
-                      ('streamrate2', int),
-                      ('heartbeatreport', int),
-                      ('radiosetup', int),
-                      ('paramretry', int),
-                      ('moddebug', int),
-                      ('rc1mul', int),
-                      ('rc2mul', int),
-                      ('rc4mul', int)]
-        self.link = 1
-        self.altreadout = 10
-        self.distreadout = 200
-        self.battreadout = 0
-        self.basealtitude = -1
-        self.heartbeat = 1
-        self.numcells = 0
-        self.mavfwd = 1
-        self.mavfwd_rate = 0
-        self.speech = 0
-        self.streamrate = 4
-        self.streamrate2 = 4
-        self.radiosetup = 0
-        self.heartbeatreport = 1
-        self.paramretry = 10
-        self.rc1mul = 1
-        self.rc2mul = 1
-        self.rc4mul = 1
-        self.moddebug = 0
-
-    def set(self, vname, value):
-        '''set a setting'''
-        for (v,t) in sorted(self.vars):
-            if v == vname:
-                try:
-                    value = t(value)
-                except:
-                    print("Unable to convert %s to type %s" % (value, t))
-                    return
-                setattr(self, vname, value)
-                return
-
-    def show(self, v):
-        '''show settings'''
-        print("%20s %s" % (v, getattr(self, v)))
-
-    def show_all(self):
-        '''show all settings'''
-        for (v,t) in sorted(self.vars):
-            self.show(v)
-
 class MPStatus(object):
     '''hold status information about the mavproxy'''
     def __init__(self):
@@ -105,6 +44,7 @@ class MPStatus(object):
         self.mav_error = 0
         self.target_system = 1
         self.target_component = 1
+        self.rallyloader = mavwp.MAVRallyLoader(self.target_system, self.target_component)
         self.speech = None
         self.altitude = 0
         self.last_altitude_announce = 0.0
@@ -119,6 +59,7 @@ class MPStatus(object):
         self.last_override = [ 0 ] * 8
         self.override_counter = 0
         self.flightmode = 'MAV'
+        self.last_mode_announce = 0
         self.logdir = None
         self.last_heartbeat = 0
         self.last_message = 0
@@ -167,6 +108,7 @@ class MPState(object):
     def __init__(self):
         self.console = textconsole.SimpleConsole()
         self.map = None
+        self.map_functions = {}
         self.settings = mp_settings.MPSettings(
             [ ('link', int, 1),
               ('altreadout', int, 10),
@@ -185,7 +127,9 @@ class MPState(object):
               ('rc2mul', int, 1),
               ('rc4mul', int, 1),
               ('shownoise', int, 1),
-              ('basealt', int, 0)]
+              ('basealt', int, 0),
+              ('wpalt', int, 100),
+              ('parambatch', int, 10)]
             )
         self.status = MPStatus()
 
@@ -401,6 +345,29 @@ def cmd_fbwa(args):
     '''set FBWA mode'''
     mpstate.master().set_mode_fbwa()
 
+def cmd_guided(args):
+    '''set GUIDED target'''
+    if len(args) != 1:
+        print("Usage: guided ALTITUDE")
+        return
+    try:
+        latlon = mpstate.map_state.click_position
+    except Exception:
+        print("No map available")
+        return
+    if latlon is None:
+        print("No map click position available")
+        return        
+    altitude = int(args[0])
+    print("Guided %s %d" % (str(latlon), altitude))
+    mpstate.master().mav.mission_item_send(mpstate.status.target_system,
+                                           mpstate.status.target_component,
+                                           0,
+                                           mavutil.mavlink.MAV_FRAME_GLOBAL_RELATIVE_ALT,
+                                           mavutil.mavlink.MAV_CMD_NAV_WAYPOINT,
+                                           2, 0, 0, 0, 0, 0,
+                                           latlon[0], latlon[1], altitude)
+
 def process_waypoint_request(m, master):
     '''process a waypoint request from the master'''
     if (not mpstate.status.loading_waypoints or
@@ -481,11 +448,70 @@ def save_waypoints(filename):
         return
     print("Saved %u waypoints to %s" % (mpstate.status.wploader.count(), filename))
 
+def wp_draw_callback(points):
+    '''callback from drawing waypoints'''
+    if len(points) < 3:
+        return
+    from MAVProxy.modules.lib import mp_util
+    home = mpstate.status.wploader.wp(0)
+    mpstate.status.wploader.clear()
+    mpstate.status.wploader.target_system = mpstate.status.target_system
+    mpstate.status.wploader.target_component = mpstate.status.target_component
+    mpstate.status.wploader.add(home)
+    for p in points:
+        mpstate.status.wploader.add_latlonalt(p[0], p[1], mpstate.settings.wpalt)
+    mpstate.master().waypoint_clear_all_send()
+    if mpstate.status.wploader.count() == 0:
+        return
+    mpstate.status.loading_waypoints = True
+    mpstate.status.loading_waypoint_lasttime = time.time()
+    mpstate.master().waypoint_count_send(mpstate.status.wploader.count())
+
+def wp_loop():
+    '''close the loop on a mission'''
+    loader = mpstate.status.wploader
+    if loader.count() < 2:
+        print("Not enough waypoints (%u)" % loader.count())
+        return
+    wp = loader.wp(loader.count()-2)
+    if wp.command == mavutil.mavlink.MAV_CMD_DO_JUMP:
+        print("Mission is already looped")
+        return
+    wp = mavutil.mavlink.MAVLink_mission_item_message(0, 0, 0, 0, mavutil.mavlink.MAV_CMD_DO_JUMP,
+                                                      0, 1, 1, -1, 0, 0, 0, 0, 0)
+    loader.add(wp)
+    loader.add(loader.wp(1))
+    mpstate.status.loading_waypoints = True
+    mpstate.status.loading_waypoint_lasttime = time.time()
+    mpstate.master().waypoint_count_send(mpstate.status.wploader.count())
+    print("Closed loop on mission")
+
+def set_home_location():
+    '''set home location from last map click'''
+    try:
+        latlon = mpstate.map_state.click_position
+    except Exception:
+        print("No map available")
+        return
+    lat = float(latlon[0])
+    lon = float(latlon[1])
+    if mpstate.status.wploader.count() == 0:
+        mpstate.status.wploader.add_latlonalt(lat, lon, 0)
+    w = mpstate.status.wploader.wp(0)
+    w.x = lat
+    w.y = lon
+    mpstate.status.wploader.set(w, 0)
+    mpstate.status.loading_waypoints = True
+    mpstate.status.loading_waypoint_lasttime = time.time()
+    mpstate.master().mav.mission_write_partial_list_send(mpstate.status.target_system,
+                                                         mpstate.status.target_component,
+                                                         0, 0)
+    
 
 def cmd_wp(args):
     '''waypoint commands'''
     if len(args) < 1:
-        print("usage: wp <list|load|save|set|clear>")
+        print("usage: wp <list|load|update|save|set|clear|loop>")
         return
 
     if args[0] == "load":
@@ -524,8 +550,22 @@ def cmd_wp(args):
         mpstate.master().waypoint_set_current_send(int(args[1]))
     elif args[0] == "clear":
         mpstate.master().waypoint_clear_all_send()
+    elif args[0] == "draw":
+        if not 'draw_lines' in mpstate.map_functions:
+            print("No map drawing available")
+            return        
+        if mpstate.status.wploader.count() == 0:
+            print("Need home location - refresh waypoints")
+            return
+        mpstate.map_functions['draw_lines'](wp_draw_callback)
+        print("Drawing waypoints on map")
+    elif args[0] == "sethome":
+        set_home_location()        
+    elif args[0] == "loop":
+        wp_loop()        
     else:
-        print("Usage: wp <list|load|save|set|show|clear>")
+        print("Usage: wp <list|load|save|set|show|clear|draw|loop>")
+
 
 
 def fetch_fence_point(i):
@@ -547,17 +587,8 @@ def fetch_fence_point(i):
 
 
 
-def load_fence(filename):
-    '''load fence points from a file'''
-    try:
-        mpstate.status.fenceloader.target_system = mpstate.status.target_system
-        mpstate.status.fenceloader.target_component = mpstate.status.target_component
-        mpstate.status.fenceloader.load(filename)
-    except Exception, msg:
-        print("Unable to load %s - %s" % (filename, msg))
-        return
-    print("Loaded %u geo-fence points from %s" % (mpstate.status.fenceloader.count(), filename))
-
+def send_fence():
+    '''send fence points from fenceloader'''
     # must disable geo-fencing when loading
     action = get_mav_param('FENCE_ACTION', mavutil.mavlink.FENCE_ACTION_NONE)
     param_set('FENCE_ACTION', mavutil.mavlink.FENCE_ACTION_NONE)
@@ -577,6 +608,18 @@ def load_fence(filename):
             return
     param_set('FENCE_ACTION', action)
 
+def load_fence(filename):
+    '''load fence points from a file'''
+    try:
+        mpstate.status.fenceloader.target_system = mpstate.status.target_system
+        mpstate.status.fenceloader.target_component = mpstate.status.target_component
+        mpstate.status.fenceloader.load(filename)
+    except Exception, msg:
+        print("Unable to load %s - %s" % (filename, msg))
+        return
+    print("Loaded %u geo-fence points from %s" % (mpstate.status.fenceloader.count(), filename))
+    send_fence()
+
 
 def list_fence(filename):
     '''list fence points, optionally saving to a file'''
@@ -609,10 +652,28 @@ def list_fence(filename):
         print("Saved fence to %s" % fencetxt)
 
 
+def fence_draw_callback(points):
+    '''callback from drawing a fence'''
+    if len(points) < 3:
+        return
+    from MAVProxy.modules.lib import mp_util
+    mpstate.status.fenceloader.clear()
+    mpstate.status.fenceloader.target_system = mpstate.status.target_system
+    mpstate.status.fenceloader.target_component = mpstate.status.target_component
+    bounds = mp_util.polygon_bounds(points)
+    (lat, lon, width, height) = bounds
+    center = (lat+width/2, lon+height/2)
+    mpstate.status.fenceloader.add_latlon(center[0], center[1])
+    for p in points:
+        mpstate.status.fenceloader.add_latlon(p[0], p[1])
+    # close it
+    mpstate.status.fenceloader.add_latlon(points[0][0], points[0][1])
+    send_fence()
+
 def cmd_fence(args):
     '''geo-fence commands'''
     if len(args) < 1:
-        print("usage: fence <list|load|save|clear>")
+        print("usage: fence <list|load|save|clear|draw>")
         return
 
     if args[0] == "load":
@@ -632,12 +693,147 @@ def cmd_fence(args):
             print("usage: fence show <filename>")
             return
         mpstate.status.fenceloader.load(args[1])
+    elif args[0] == "draw":
+        if not 'draw_lines' in mpstate.map_functions:
+            print("No map drawing available")
+            return        
+        mpstate.map_functions['draw_lines'](fence_draw_callback)
+        print("Drawing fence on map")
     elif args[0] == "clear":
         param_set('FENCE_TOTAL', 0)
     else:
-        print("Usage: fence <list|load|save|show|clear>")
+        print("Usage: fence <list|load|save|show|clear|draw>")
 
 
+def fetch_rally_point(i):
+    '''fetch one rally point'''
+    mpstate.master().mav.rally_fetch_point_send(mpstate.status.target_system,
+                                                mpstate.status.target_component, i)
+    tstart = time.time()
+    p = None
+    while time.time() - tstart < 1:
+        p = mpstate.master().recv_match(type='RALLY_POINT', blocking=False)
+        if p is not None:
+            break
+        time.sleep(0.1)
+        continue
+    if p is None:
+        mpstate.console.error("Failed to fetch rally point %u" % i)
+        return None
+    return p
+
+def send_rally_points():
+    '''send rally points from fenceloader'''
+    param_set('RALLY_TOTAL', mpstate.status.rallyloader.rally_count())
+    for i in range(mpstate.status.rallyloader.rally_count()):
+        p = mpstate.status.rallyloader.rally_point(i)
+        mpstate.master().mav.send(p)
+        
+        #Don't know why this check was being done, but not discounting yet...
+        #Will remove before committing if I don't discover why.
+        #p2 = fetch_rally_point(i)
+        #
+        #if (p.idx != p2.idx or
+        #    abs(p.lat - p2.lat) >= 0.00003 or
+        #    abs(p.lng - p2.lng) >= 0.00003):
+        #    print("Failed to send fence point %u" % i)
+        #    param_set('FENCE_ACTION', action)
+        #    return
+
+    #TODO rally land points
+
+def list_rally_points():
+    mpstate.status.rallyloader.clear()
+    rally_count = get_mav_param('RALLY_TOTAL', 0)
+    if rally_count == 0:
+        print("No rally points")
+        return
+    for i in range(int(rally_count)):
+        p = fetch_rally_point(i)
+        if p is None:
+            return
+        mpstate.status.rallyloader.append_rally_point(p)
+
+    #TODO: rally_land points
+
+    for i in range(mpstate.status.rallyloader.rally_count()):
+        p = mpstate.status.rallyloader.rally_point(i)
+        mpstate.console.writeln("lat=%f lng=%f alt=%f break_alt=%f land_dir=%f" % (p.lat * 1e-7, p.lng * 1e-7, p.alt, p.break_alt, p.land_dir))
+
+    #TODO: rally_land points
+
+def cmd_rally(args):
+    '''rally point commands'''
+    #TODO: add_land arg
+    if(len(args) < 1):
+        print("Usage: rally <list|load|save|add|clear>")
+        return
+
+    elif(args[0] == "add"):
+        if (len(args) < 2):
+            print("Usage: rally add ALT <BREAK_ALT> <LAND_HDG>")
+            return
+
+        if (mpstate.status.rallyloader.rally_count() > 4):
+            print ("Only 5 rally points possible per flight plan.")
+            return
+
+        try:
+            latlon = mpstate.map_state.click_position
+        except Exception:
+            print("No map available")
+            return
+        if latlon is None:
+            print("No map click position available")
+            return
+
+        alt = float(args[1]);
+        break_alt = 0.0
+        land_hdg = 0.0;
+        if (len(args) > 2):
+            break_alt = float(args[2]);
+        if (len(args) > 3):
+            land_hdg = float(args[3]);
+
+        mpstate.status.rallyloader.create_and_append_rally_point(latlon[0] * 1e7, latlon[1] * 1e7, alt, break_alt, land_hdg, 0)
+
+        send_rally_points();
+
+        print("Added Rally point at %s %f" % (str(latlon), alt))
+    
+    elif args[0] == "clear":
+        mpstate.status.rallyloader.clear()
+        param_set('RALLY_TOTAL', 0)
+
+    elif(args[0] == "list"):
+        list_rally_points()
+
+    elif(args[0] == "load"):
+        if (len(args) < 2):
+            print("Usage: rally load filename")
+
+        try:
+            mpstate.status.rallyloader.load(args[1])
+        except Exception, msg:
+            print("Unable to load %s - %s" % (args[1], msg))
+            return
+    
+        send_rally_points()
+
+        print("Loaded %u rally points from %s" % (mpstate.status.rallyloader.rally_count(), args[1]))
+
+    elif(args[0] == "save"):
+        if (len(args) < 2):
+            print("Usage: rally save filename");
+
+        mpstate.status.rallyloader.save(args[1]);
+
+        print "Saved rally file ", args[1];
+
+    else:
+        #TODO: add_land arg
+        print("Usage: rally <list|load|save|add|clear>")
+
 def param_set(name, value, retries=3):
     '''set a parameter'''
     name = name.upper()
@@ -725,6 +921,95 @@ def cmd_set(args):
     else:
         mpstate.settings.set(args[0], args[1])
 
+tune_options = {
+    'None':             '0',
+    'StabRollPitchkP':  '1',
+    'RateRollPitchkP':  '4',
+    'RateRollPitchkI':  '6',
+    'RateRollPitchkD':  '21',
+    'StabYawkP':        '3',
+    'RateYawkP':        '6',
+    'RateYawkD':        '26',
+    'AltitudeHoldkP':   '14',
+    'ThrottleRatekP':   '7',
+    'ThrottleRatekD':   '37',
+    'ThrottleAccelkP':  '34',
+    'ThrottleAccelkI':  '35',
+    'ThrottleAccelkD':  '36',
+    'LoiterPoskP':      '12',
+    'LoiterRatekP':     '22',
+    'LoiterRatekI':     '28',
+    'LoiterRatekD':     '29',
+    'WPSpeed':          '10',
+    'AcroRollPitch kP': '25',
+    'AcroYawkP':        '40',
+    'RelayOnOff':       '9',
+    'HeliExtGyro':      '13',
+    'OFLoiterkP':       '17',
+    'OFLoiterkI':       '18',
+    'OFLoiterkD':       '19',
+    'AHRSYawkP':        '30',
+    'AHRSkP':           '31',
+    'INAV_TC':          '32',
+    'Declination':      '38',
+    'CircleRate':       '39',
+    'SonarGain':       '41',
+}
+
+def tune_show():
+    opt_num = str(int(get_mav_param('TUNE')))
+    option = None
+    for k in tune_options.keys():
+        if opt_num == tune_options[k]:
+            option = k
+            break
+    else:
+        print("TUNE is currently set to unknown value " + opt_num)
+        return
+    low = get_mav_param('TUNE_LOW')
+    high = get_mav_param('TUNE_HIGH')
+    print("TUNE is currently set to %s LOW=%f HIGH=%f" % (option, low/1000, high/1000))
+
+def tune_option_validate(option):
+    for k in tune_options:
+        if option.upper() == k.upper():
+            return k
+    return None
+
+# TODO: Check/show the limits of LOW and HIGH
+def cmd_tuneopt(args):
+    '''Select option for Tune Pot on Channel 6 (quadcopter only)'''
+    usage = "usage: tuneopt <set|show|reset|list>"
+    if not opts.quadcopter or 'TUNE' not in mpstate.mav_param:
+        print("This command is only available for quadcopter")
+        return
+    if len(args) < 1:
+        print(usage)
+        return
+    if args[0].lower() == 'reset':
+        param_set('TUNE', '0')
+    elif args[0].lower() == 'set':
+        if len(args) < 4:
+            print('Usage: tuneopt set OPTION LOW HIGH')
+            return
+        option = tune_option_validate(args[1])
+        if not option:
+            print('Invalid Tune option: ' + args[1])
+            return
+        low = args[2]
+        high = args[3]
+        param_set('TUNE', tune_options[option])
+        param_set('TUNE_LOW', float(low) * 1000)
+        param_set('TUNE_HIGH', float(high) * 1000)
+    elif args[0].lower() == 'show':
+        tune_show()
+    elif args[0].lower() == 'list':
+        print("Options available:")
+        for s in sorted(tune_options.keys()):
+            print('  ' + s)
+    else:
+        print(usage)
+
 def cmd_status(args):
     '''show status'''
     if len(args) == 0:
@@ -760,6 +1045,87 @@ def cmd_up(args):
     print("Adjusting TRIM_PITCH_CD from %d to %d" % (old_trim, new_trim))
     param_set('TRIM_PITCH_CD', new_trim)
 
+aux_options = {
+    "Nothing":"0",
+    "Flip":"2",
+    "SimpleMode":"3",
+    "RTL":"4",
+    "SaveTrim":"5",
+    "SaveWP":"7",
+    "MultiMode":"8",
+    "CameraTrigger":"9",
+    "Sonar":"10",
+    "Fence":"11",
+    "ResetYaw":"12",
+    "SuperSimpleMode":"13",
+    "AcroTrainer":"14",
+    "Auto":"16",
+    "AutoTune":"17",
+    "Land":"18"
+}
+
+def aux_show(channel):
+    param = "CH%s_OPT" % channel
+    opt_num = str(int(get_mav_param(param)))
+    option = None
+    for k in aux_options.keys():
+        if opt_num == aux_options[k]:
+            option = k
+            break
+    else:
+        print("AUX Channel is currently set to unknown value " + opt_num)
+        return
+    print("AUX Channel is currently set to " + option)
+
+def aux_option_validate(option):
+    for k in aux_options:
+        if option.upper() == k.upper():
+            return k
+    return None
+
+def cmd_auxopt(args):
+    '''handle AUX switches (CH7, CH8) settings'''
+    if not opts.quadcopter or 'CH7_OPT' not in mpstate.mav_param:
+        print("This command is only available for quadcopter")
+        return
+    if len(args) == 0 or args[0] not in ('set', 'show', 'reset', 'list'):
+        print("Usage: auxopt set|show|reset|list")
+        return
+    if args[0] == 'list':
+        print("Options available:")
+        for s in sorted(aux_options.keys()):
+            print('  ' + s)
+    elif args[0] == 'show':
+        if len(args) > 2 and args[1] not in ['7', '8', 'all']:
+            print("Usage: auxopt show [7|8|all]")
+            return
+        if len(args) < 2 or args[1] == 'all':
+            aux_show('7')
+            aux_show('8')
+            return
+        aux_show(args[1])
+    elif args[0] == 'reset':
+        if len(args) < 2 or args[1] not in ['7', '8', 'all']:
+            print("Usage: auxopt reset 7|8|all")
+            return
+        if args[1] == 'all':
+            param_set('CH7_OPT', '0')
+            param_set('CH8_OPT', '0')
+            return
+        param = "CH%s_OPT" % args[1]
+        param_set(param, '0')
+    elif args[0] == 'set':
+        if len(args) < 3 or args[1] not in ['7', '8']:
+            print("Usage: auxopt set 7|8 OPTION")
+            return
+        option = aux_option_validate(args[2])
+        if not option:
+            print("Invalid option " + args[2])
+            return
+        param = "CH%s_OPT" % args[1]
+        param_set(param, aux_options[option])
+    else:
+        print("Usage: auxopt set|show|list")
 
 def cmd_setup(args):
     mpstate.status.setup_mode = True
@@ -804,16 +1170,19 @@ def cmd_module(args):
             print("usage: module load <name>")
             return
         modname = args[1]
-        modpath = 'MAVProxy.modules.mavproxy_%s' % (modname,)
-        try:
-            m = import_package(modpath)
-            if m in mpstate.modules:
-                raise RuntimeError("module %s already loaded" % (modname,))
-            m.init(mpstate)
-            mpstate.modules.append(m)
-            print("Loaded module %s" % (modname,))
-        except Exception, msg:
-            print("Unable to load module %s: %s" % (modname, msg))
+        modpaths = ['MAVProxy.modules.mavproxy_%s' % modname, modname]
+        for modpath in modpaths:
+            try:
+                m = import_package(modpath)
+                if m in mpstate.modules:
+                    raise RuntimeError("module %s already loaded" % (modname,))
+                m.init(mpstate)
+                mpstate.modules.append(m)
+                print("Loaded module %s" % (modname,))
+                return
+            except ImportError, msg:
+                ex = msg
+        print("Failed to load module: %s" % ex)
     elif args[0] == "reload":
         if len(args) < 2:
             print("usage: module reload <name>")
@@ -881,6 +1250,14 @@ def cmd_arm(args):
   '''arm motors'''
   mpstate.master().arducopter_arm()
 
+def cmd_time(args):
+  '''show autopilot time'''
+  tusec = mpstate.master().field('SYSTEM_TIME', 'time_unix_usec', 0)
+  if tusec == 0:
+      print("No SYSTEM_TIME time available")
+      return
+  print("%s (%s)\n" % (time.ctime(tusec * 1.0e-6), time.ctime()))
+
 
 def cmd_disarm(args):
   '''disarm motors'''
@@ -904,7 +1281,9 @@ command_map = {
     'rc'      : (cmd_rc,       'override a RC channel value'),
     'wp'      : (cmd_wp,       'waypoint management'),
     'fence'   : (cmd_fence,    'geo-fence management'),
+    'rally'   : (cmd_rally,    'rally point management'),
     'param'   : (cmd_param,    'manage APM parameters'),
+    'tuneopt' : (cmd_tuneopt,  'Select option for Tune Pot on Channel 6 (quadcopter only)'),
     'setup'   : (cmd_setup,    'go into setup mode'),
     'reset'   : (cmd_reset,    'reopen the connection to the MAVLink master'),
     'status'  : (cmd_status,   'show status'),
@@ -918,6 +1297,7 @@ command_map = {
     'rtl'     : (cmd_rtl,      'set RTL mode'),
     'manual'  : (cmd_manual,   'set MANUAL mode'),
     'fbwa'    : (cmd_fbwa,     'set FBWA mode'),
+    'guided'  : (cmd_guided,   'set GUIDED target'),
     'set'     : (cmd_set,      'mavproxy settings'),
     'bat'     : (cmd_bat,      'show battery levels'),
     'alt'     : (cmd_alt,      'show relative altitude'),
@@ -925,11 +1305,13 @@ command_map = {
     'servo'   : (cmd_servo,    'set a servo value'),
     'reboot'  : (cmd_reboot,   'reboot the autopilot'),
     'up'      : (cmd_up,       'adjust TRIM_PITCH_CD up by 5 degrees'),
+    'auxopt'  : (cmd_auxopt,   'select option for aux switches on CH7 and CH8 (ArduCopter only)'),
     'watch'   : (cmd_watch,    'watch a MAVLink pattern'),
     'module'  : (cmd_module,   'module commands'),
     'alias'   : (cmd_alias,    'command aliases'),
-    'arm'     : (cmd_arm,      'ArduCopter arm motors'),
-    'disarm'  : (cmd_disarm,   'ArduCopter disarm motors')
+    'arm'     : (cmd_arm,      'Copter/Plane arm motors'),
+    'time'    : (cmd_time,     'Show autopilot time'),
+    'disarm'  : (cmd_disarm,   'Copter/Plane disarm motors')
     }
 
 def process_stdin(line):
@@ -945,10 +1327,11 @@ def process_stdin(line):
             mpstate.status.flightmode = "MAV"
             mpstate.rl.set_prompt("MAV> ")
             return
-        if line == '+++':
-            mpstate.master().write(line)
-        else:
-            mpstate.master().write(line + '\r')
+        if line != '+++':
+            line += '\r'
+        for c in line:
+            time.sleep(0.01)
+            mpstate.master().write(c)
         return
 
     if not line:
@@ -1053,7 +1436,7 @@ def vcell_to_battery_percent(vcell):
     elif vcell > 3.81:
         # 3.81 is 17% remaining, from flight logs
         return 17.0 + 83.0 * (vcell - 3.81) / (4.1 - 3.81)
-    elif vcell > 3.81:
+    elif vcell > 3.2:
         # below 3.2 it degrades fast. It's dead at 3.2
         return 0.0 + 17.0 * (vcell - 3.20) / (3.81 - 3.20)
     # it's dead or disconnected
@@ -1112,6 +1495,11 @@ def battery_report():
 
 def handle_msec_timestamp(m, master):
     '''special handling for MAVLink packets with a time_boot_ms field'''
+
+    if m.get_type() == 'GLOBAL_POSITION_INT':
+        # this is fix time, not boot time
+        return
+
     msec = m.time_boot_ms
     if msec + 30000 < master.highest_msec:
         say('Time has wrapped')
@@ -1144,9 +1532,10 @@ def report_altitude(altitude):
     if (int(mpstate.settings.altreadout) > 0 and
         math.fabs(mpstate.status.altitude - mpstate.status.last_altitude_announce) >= int(mpstate.settings.altreadout)):
         mpstate.status.last_altitude_announce = mpstate.status.altitude
-        rounded_alt = int(mpstate.settings.altreadout) * ((5+int(mpstate.status.altitude)) / int(mpstate.settings.altreadout))
+        rounded_alt = int(mpstate.settings.altreadout) * ((mpstate.settings.altreadout/2 + int(mpstate.status.altitude)) / int(mpstate.settings.altreadout))
         say("height %u" % rounded_alt, priority='notification')
 
+
 def master_send_callback(m, master):
     '''called on sending a message'''
     mtype = m.get_type()
@@ -1171,7 +1560,7 @@ def master_callback(m, master):
     mtype = m.get_type()
 
     # and log them
-    if mtype != 'BAD_DATA' and mpstate.logqueue:
+    if mtype not in ['BAD_DATA','LOG_DATA'] and mpstate.logqueue:
         # put link number in bottom 2 bits, so we can analyse packet
         # delay in saved logs
         usec = get_usec()
@@ -1192,7 +1581,7 @@ def master_callback(m, master):
                       'NAV_CONTROLLER_OUTPUT' ]:
             return
 
-    if mtype == 'HEARTBEAT':
+    if mtype == 'HEARTBEAT' and m.get_srcSystem() != 255:
         if (mpstate.status.target_system != m.get_srcSystem() or
             mpstate.status.target_component != m.get_srcComponent()):
             mpstate.status.target_system = m.get_srcSystem()
@@ -1303,8 +1692,9 @@ def master_callback(m, master):
 
     elif mtype == "SYS_STATUS":
         battery_update(m)
-        if master.flightmode != mpstate.status.flightmode:
+        if master.flightmode != mpstate.status.flightmode and time.time() > mpstate.status.last_mode_announce + 2:
             mpstate.status.flightmode = master.flightmode
+            mpstate.status.last_mode_announce = time.time()
             mpstate.rl.set_prompt(mpstate.status.flightmode + "> ")
             say("Mode " + mpstate.status.flightmode)
 
@@ -1408,7 +1798,7 @@ def master_callback(m, master):
 def process_master(m):
     '''process packets from the MAVLink master'''
     try:
-        s = m.recv()
+        s = m.recv(16*1024)
     except Exception:
         return
     if mpstate.logqueue_raw:
@@ -1547,6 +1937,15 @@ def check_link_status():
             say("link %u down" % (master.linknum+1))
             master.linkerror = True
 
+def send_heartbeat(master):
+    if master.mavlink10():
+        master.mav.heartbeat_send(mavutil.mavlink.MAV_TYPE_GCS, mavutil.mavlink.MAV_AUTOPILOT_INVALID,
+                                  0, 0, 0)
+    else:
+        MAV_GROUND = 5
+        MAV_AUTOPILOT_NONE = 4
+        master.mav.heartbeat_send(MAV_GROUND, MAV_AUTOPILOT_NONE)
+
 def periodic_tasks():
     '''run periodic checks'''
     if mpstate.status.setup_mode:
@@ -1558,13 +1957,7 @@ def periodic_tasks():
     if heartbeat_period.trigger() and mpstate.settings.heartbeat != 0:
         mpstate.status.counters['MasterOut'] += 1
         for master in mpstate.mav_master:
-            if master.mavlink10():
-                master.mav.heartbeat_send(mavutil.mavlink.MAV_TYPE_GCS, mavutil.mavlink.MAV_AUTOPILOT_INVALID,
-                                          0, 0, 0)
-            else:
-                MAV_GROUND = 5
-                MAV_AUTOPILOT_NONE = 4
-                master.mav.heartbeat_send(MAV_GROUND, MAV_AUTOPILOT_NONE)
+            send_heartbeat(master)
 
     if heartbeat_check_period.trigger():
         check_link_status()
@@ -1577,9 +1970,11 @@ def periodic_tasks():
         elif mpstate.mav_param_count != 0 and len(mpstate.mav_param_set) != mpstate.mav_param_count:
             if mpstate.master().time_since('PARAM_VALUE') >= 1:
                 diff = set(range(mpstate.mav_param_count)).difference(mpstate.mav_param_set)
-                if len(diff) > 0:
+                count = 0
+                while len(diff) > 0 and count < mpstate.settings.parambatch:
                     idx = diff.pop()
                     mpstate.master().param_fetch_one(idx)
+                    count += 1
 
         # cope with packet loss fetching mission
         if mpstate.master().time_since('MISSION_ITEM') >= 2 and mpstate.status.wploader.count() < getattr(mpstate.status.wploader,'expected_count',0):
@@ -1619,6 +2014,7 @@ def main_loop():
     '''main processing loop'''
     if not mpstate.status.setup_mode and not opts.nowait:
         for master in mpstate.mav_master:
+            send_heartbeat(master)
             master.wait_heartbeat()
             if len(mpstate.mav_param) < 10 or not mpstate.continue_mode:
                 mpstate.mav_param_set = set()
@@ -1648,13 +2044,13 @@ def main_loop():
         for m in mpstate.mav_outputs:
             rin.append(m.fd)
         if rin == []:
-            time.sleep(0.001)
+            time.sleep(0.0001)
             continue
 
         for fd in mpstate.select_extra:
             rin.append(fd)
         try:
-            (rin, win, xin) = select.select(rin, [], [], 0.001)
+            (rin, win, xin) = select.select(rin, [], [], 0.0001)
         except select.error:
             continue
 
@@ -1766,12 +2162,14 @@ if __name__ == '__main__':
     parser.add_option("--auto-protocol", action='store_true', default=False, help="Auto detect MAVLink protocol version")
     parser.add_option("--nowait", action='store_true', default=False, help="don't wait for HEARTBEAT on startup")
     parser.add_option("--continue", dest='continue_mode', action='store_true', default=False, help="continue logs")
+    parser.add_option("--dialect",  default="ardupilotmega", help="MAVLink dialect")
 
     (opts, args) = parser.parse_args()
 
     if opts.mav09:
         os.environ['MAVLINK09'] = '1'
     from pymavlink import mavutil, mavwp, mavparm
+    mavutil.set_dialect(opts.dialect)
 
     # global mavproxy state
     mpstate = MPState()
@@ -1810,12 +2208,7 @@ Auto-detected serial ports are:
 
     # open master link
     for mdev in opts.master:
-        if mdev.startswith('tcp:'):
-            m = mavutil.mavtcp(mdev[4:])
-        elif mdev.find(':') != -1:
-            m = mavutil.mavudp(mdev, input=True)
-        else:
-            m = mavutil.mavserial(mdev, baud=opts.baudrate, autoreconnect=True)
+        m = mavutil.mavlink_connection(mdev, autoreconnect=True, baud=opts.baudrate)
         m.mav.set_callback(master_callback, m)
         if hasattr(m.mav, 'set_send_callback'):
             m.mav.set_send_callback(master_send_callback, m)
@@ -1860,6 +2253,7 @@ Auto-detected serial ports are:
 
     msg_period = mavutil.periodic_event(1.0/15)
     param_period = mavutil.periodic_event(1)
+    log_period = mavutil.periodic_event(2)
     heartbeat_period = mavutil.periodic_event(1)
     battery_period = mavutil.periodic_event(0.1)
     if mpstate.sitl_output:
@@ -1884,6 +2278,12 @@ Auto-detected serial ports are:
         else:
             print("no script %s" % start_script)
 
+    if not opts.setup:
+        # some core functionality is in modules
+        standard_modules = ['log']
+        for m in standard_modules:
+            process_stdin('module load %s' % m)
+
     if opts.console:
         process_stdin('module load console')
 
